==> osm2go_0.1-1.tar.gz: osm2go-0.1/info.c <==
/*
 * Copyright (C) 2008 Till Harbaum <till@harbaum.org>.
 *
 * This file is part of OSM2Go.
 *
 * OSM2Go is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OSM2Go is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OSM2Go.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "appdata.h"

typedef struct {
  appdata_t *appdata;
  GtkWidget *dialog, *view;
  GtkListStore *store;
  GtkWidget *but_add, *but_remove, *but_edit;
  tag_t **tag;
} tag_context_t;

enum {
  TAG_COL_KEY = 0,
  TAG_COL_VALUE,
  TAG_COL_DATA,
  TAG_NUM_COLS
};

static gboolean
view_selection_func(GtkTreeSelection *selection, GtkTreeModel *model,
		     GtkTreePath *path, gboolean path_currently_selected,
		     gpointer userdata) {
  tag_context_t *context = (tag_context_t*)userdata;
  GtkTreeIter iter;
    
  if(gtk_tree_model_get_iter(model, &iter, path)) {
    g_assert(gtk_tree_path_get_depth(path) == 1);

    if(!path_currently_selected) {
      tag_t *tag;
      gtk_tree_model_get(model, &iter, TAG_COL_DATA, &tag, -1);

      /* you just cannot delete or edit the "created_by" tag */
      if(strcasecmp(tag->key, "created_by") == 0) {
	gtk_widget_set_sensitive(context->but_remove, FALSE);
	gtk_widget_set_sensitive(context->but_edit, FALSE);
      } else {
	gtk_widget_set_sensitive(context->but_remove, TRUE);
	gtk_widget_set_sensitive(context->but_edit, TRUE);
      }
    }
  }
  
  return TRUE; /* allow selection state to change */
}

static void on_tag_remove(GtkWidget *but, tag_context_t *context) {
  GtkTreeSelection *selection;
  GtkTreeModel     *model;
  GtkTreeIter       iter;

  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(context->view));
  if(gtk_tree_selection_get_selected(selection, &model, &iter)) {
    tag_t *tag;
    gtk_tree_model_get(model, &iter, TAG_COL_DATA, &tag, -1);

    g_assert(tag);

    /* de-chain */
    printf("de-chaining tag %s/%s\n", tag->key, tag->value);
    tag_t **prev = context->tag;
    while(*prev != tag) prev = &((*prev)->next);
    *prev = tag->next;

    /* free tag itself */
    gpx_tag_free(tag);

    /* and remove from store */
    gtk_list_store_remove(GTK_LIST_STORE(model), &iter);

    /* but is NULL if remove has been called due to a cancelled add */
    if(but) {
      /* something has been removed -> now dirty */
      map_item_set_flags(context->appdata->map->selected, GPX_FLAG_DIRTY, 0);
    }
  }
  
  /* disable remove and edit buttons */
  gtk_widget_set_sensitive(context->but_remove,  FALSE);
  gtk_widget_set_sensitive(context->but_edit,  FALSE);
}

static gboolean tag_edit(tag_context_t *context) {

  GtkTreeModel *model;
  GtkTreeIter iter;
  tag_t *tag;

  GtkTreeSelection *sel = gtk_tree_view_get_selection(
		      GTK_TREE_VIEW(context->view));
  if(!sel) {
    printf("got no selection object\n");
    return FALSE;
  }

  if(!gtk_tree_selection_get_selected(sel, &model, &iter)) {
    printf("nothing selected\n");
    return FALSE;
  }
  
  gtk_tree_model_get(model, &iter, TAG_COL_DATA, &tag, -1);
  printf("got %s/%s\n", tag->key, tag->value);

  GtkWidget *dialog = gtk_dialog_new_with_buttons(_("Edit Tag"),
	  GTK_WINDOW(context->dialog), GTK_DIALOG_MODAL,
          GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
	  GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);

  GtkWidget *label, *key, *value;
  GtkWidget *table = gtk_table_new(2, 2, FALSE);

  gtk_table_attach_defaults(GTK_TABLE(table),
		    label = gtk_label_new(_("Key:")), 0, 1, 0, 1);
  gtk_misc_set_alignment(GTK_MISC(label), 1.f, 0.5f);
  gtk_table_attach_defaults(GTK_TABLE(table), 
		    key = gtk_entry_new(), 1, 2, 0, 1);

  gtk_table_attach_defaults(GTK_TABLE(table),
		    label = gtk_label_new(_("Value:")), 0, 1, 1, 2);
  gtk_misc_set_alignment(GTK_MISC(label), 1.f, 0.5f);
  gtk_table_attach_defaults(GTK_TABLE(table), 
		    value = gtk_entry_new(), 1, 2, 1, 2);

  gtk_entry_set_text(GTK_ENTRY(key), tag->key);
  gtk_entry_set_text(GTK_ENTRY(value), tag->value);

  gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(dialog)->vbox), table);

  gtk_widget_show_all(dialog);

  if(GTK_RESPONSE_ACCEPT == gtk_dialog_run(GTK_DIALOG(dialog))) {
    free(tag->key); free(tag->value);
    tag->key = strdup((char*)gtk_entry_get_text(GTK_ENTRY(key)));
    tag->value = strdup((char*)gtk_entry_get_text(GTK_ENTRY(value)));
    printf("setting %s/%s\n", tag->key, tag->value);

    gtk_list_store_set(context->store, &iter,
		       TAG_COL_KEY, tag->key,
		       TAG_COL_VALUE, tag->value,
		       -1);

    gtk_widget_destroy(dialog);
    return TRUE;
  }

  gtk_widget_destroy(dialog);
  return FALSE;
}

static void on_tag_edit(GtkWidget *button, tag_context_t *context) {
  if(tag_edit(context))
    map_item_set_flags(context->appdata->map->selected, GPX_FLAG_DIRTY, 0);
}

static void on_tag_add(GtkWidget *button, tag_context_t *context) {
  /* search end of tag chain */
  tag_t **tag = context->tag;
  while(*tag) 
    tag = &(*tag)->next;

  /* create and append a new tag */
  *tag = g_new0(tag_t, 1);
  if(!*tag) {
    errorf(GTK_WIDGET(context->appdata->window), _("Out of memory"));
    return;
  }

  /* fill with some dummy data */
  (*tag)->key = strdup(_("<key>"));
  (*tag)->value = strdup(_("<value>"));

  /* append a row for the new data */
  GtkTreeIter iter;
  gtk_list_store_append(context->store, &iter);
  gtk_list_store_set(context->store, &iter,
		     TAG_COL_KEY, (*tag)->key,
		     TAG_COL_VALUE, (*tag)->value,
		     TAG_COL_DATA, *tag,
		     -1);

  gtk_tree_selection_select_iter(gtk_tree_view_get_selection(
		     GTK_TREE_VIEW(context->view)), &iter);

  if(!tag_edit(context)) {
    printf("cancelled\n");
    on_tag_remove(NULL, context);
  } else
    map_item_set_flags(context->appdata->map->selected, 
		       GPX_FLAG_DIRTY, 0);
}

static GtkWidget *tag_widget(tag_context_t *context) {
  GtkWidget *vbox = gtk_vbox_new(FALSE,3);
  context->view = gtk_tree_view_new();

  gtk_tree_selection_set_select_function(
	 gtk_tree_view_get_selection(GTK_TREE_VIEW(context->view)), 
	 view_selection_func, 
	 context, NULL);

  /* --- "Key" column --- */
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(context->view),
	-1, _("Key"), renderer, "text", TAG_COL_KEY, NULL);

  /* --- "Value" column --- */
  renderer = gtk_cell_renderer_text_new();
  g_object_set(renderer, "ellipsize", PANGO_ELLIPSIZE_END, NULL );
  GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes(
		 _("Value"), renderer, "text", TAG_COL_VALUE, NULL);
  gtk_tree_view_column_set_expand(column, TRUE);
  gtk_tree_view_insert_column(GTK_TREE_VIEW(context->view), column, -1);

  /* build and fill the store */
  context->store = gtk_list_store_new(TAG_NUM_COLS, 
		G_TYPE_STRING, G_TYPE_STRING, G_TYPE_POINTER);

  gtk_tree_view_set_model(GTK_TREE_VIEW(context->view), 
			  GTK_TREE_MODEL(context->store));

  GtkTreeIter iter;
  tag_t *tag = *context->tag;
  while(tag) {
    /* Append a row and fill in some data */
    gtk_list_store_append(context->store, &iter);
    gtk_list_store_set(context->store, &iter,
		       TAG_COL_KEY, tag->key,
		       TAG_COL_VALUE, tag->value,
		       TAG_COL_DATA, tag,
		       -1);
    tag = tag->next;
  }
  
  g_object_unref(context->store);

  /* put it into a scrolled window */
  GtkWidget *scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window), 
				 GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolled_window), 
				      GTK_SHADOW_ETCHED_IN);
  //  gtk_container_set_border_width(GTK_CONTAINER(scrolled_window), 3);
  gtk_container_add(GTK_CONTAINER(scrolled_window), context->view);
  gtk_box_pack_start_defaults(GTK_BOX(vbox), scrolled_window);

  /* ------- button box ------------ */

  GtkWidget *hbox = gtk_hbox_new(TRUE,3);
  context->but_add = gtk_button_new_with_label(_("Add..."));
  gtk_box_pack_start_defaults(GTK_BOX(hbox), context->but_add);
  gtk_signal_connect(GTK_OBJECT(context->but_add), "clicked",
  		     GTK_SIGNAL_FUNC(on_tag_add), context);

  context->but_edit = gtk_button_new_with_label(_("Edit..."));
  gtk_widget_set_sensitive(context->but_edit, FALSE);
  gtk_box_pack_start_defaults(GTK_BOX(hbox), context->but_edit);
  gtk_signal_connect(GTK_OBJECT(context->but_edit), "clicked",
  		     GTK_SIGNAL_FUNC(on_tag_edit), context);

  context->but_remove = gtk_button_new_with_label(_("Remove"));
  gtk_widget_set_sensitive(context->but_remove, FALSE);
  gtk_box_pack_start_defaults(GTK_BOX(hbox), context->but_remove);
  gtk_signal_connect(GTK_OBJECT(context->but_remove), "clicked",
		     GTK_SIGNAL_FUNC(on_tag_remove), context);

  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
  return vbox;
}

void info_dialog(appdata_t *appdata) {
  g_assert(appdata->map->selected);
  tag_context_t *context = g_new0(tag_context_t, 1);
  user_t *user = NULL;
  char *str = NULL;
  time_t stime = 0;

  context->appdata = appdata;

  switch(appdata->map->selected->type) {
  case MAP_TYPE_NODE:
    str = g_strdup_printf(_("Node #%ld"), appdata->map->selected->node->id);
    user = appdata->map->selected->node->user;
    context->tag = &appdata->map->selected->node->tag;
    stime = appdata->map->selected->node->time;
    break;
  case MAP_TYPE_WAY:
    str = g_strdup_printf(_("Way #%ld"), appdata->map->selected->way->id);
    user = appdata->map->selected->way->user;
    context->tag = &appdata->map->selected->way->tag;
    stime = appdata->map->selected->way->time;
    break;
  default:
    g_assert((appdata->map->selected->type == MAP_TYPE_NODE) ||
	     (appdata->map->selected->type == MAP_TYPE_WAY));
    break;
  }

  context->dialog = gtk_dialog_new_with_buttons(str,
	  GTK_WINDOW(appdata->window), GTK_DIALOG_MODAL,
	  GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE, NULL);
  g_free(str);

  /* making the dialog a little wider makes it less "crowded" */
#ifdef USE_HILDON
  gtk_window_set_default_size(GTK_WINDOW(context->dialog), 500, 300);
#else
  gtk_window_set_default_size(GTK_WINDOW(context->dialog), 400, 200);
#endif

  GtkWidget *label;
  GtkWidget *table = gtk_table_new(2, 2, FALSE);  // x, y

  /* ------------ user ----------------- */
  label = gtk_label_new(_("User:"));
  gtk_table_attach_defaults(GTK_TABLE(table),  label, 0, 1, 0, 1);
  label = gtk_label_new(user->name);
  gtk_table_attach_defaults(GTK_TABLE(table),  label, 1, 2, 0, 1);

  /* ------------ time ----------------- */
  label = gtk_label_new(_("Time:"));
  gtk_table_attach_defaults(GTK_TABLE(table),  label, 0, 1, 1, 2);
  struct tm *loctime = localtime(&stime);
  char time_str[32];
  strftime(time_str, sizeof(time_str), "%x %X", loctime);
  label = gtk_label_new(time_str);
  gtk_table_attach_defaults(GTK_TABLE(table),  label, 1, 2, 1, 2);

  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(context->dialog)->vbox), table, 
		     FALSE, FALSE, 0);

  /* ------------ tags ----------------- */

  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(context->dialog)->vbox),
		     tag_widget(context), TRUE, TRUE, 0);

  /* ----------------------------------- */

  gtk_widget_show_all(context->dialog);
  
  gtk_dialog_run(GTK_DIALOG(context->dialog));

  gtk_widget_destroy(context->dialog);

  /* since nodes being parts of ways but with no tags are invisible, */
  /* the result of editing them may have changed their visibility */
  map_item_t *map_item = appdata->map->selected;
  if((map_item->type == MAP_TYPE_NODE) && (map_item->node->ways)) {
    node_t *node = map_item->node;

    printf("tag editor has been run for a node being part of a way\n");
    printf("  -> check visibility\n");

    if(!node->map_item_chain && gpx_node_has_tag(node)) {
      printf("Node has just become visible\n");

      map_item_deselect(appdata);
      map_show_node(appdata->map, node);

      g_assert(node->map_item_chain);
      map_item_select(appdata, node->map_item_chain->map_item);
      
    } else if(node->map_item_chain && !gpx_node_has_tag(node)) {

      printf("Node has just become invisible\n");
      map_item_deselect(appdata);
      map_item_chain_t *chain = node->map_item_chain;
      while(chain) {
	map_item_chain_t *next = chain->next;
	gtk_object_destroy(GTK_OBJECT(chain->map_item->item));
	chain = next;
      }

      /* the node itself is now invisible */
      g_assert(!node->map_item_chain);
      map_item_t *new = g_new0(map_item_t, 1);
      new->type = MAP_TYPE_TMP_NODE;
      new->node = node;
      new->item = NULL;
      new->highlight = FALSE;	  

      map_item_select(appdata, new);
    }
  }
}

==> osm2go_0.2-1.tar.gz: osm2go-0.2/info.c <==
/*
 * Copyright (C) 2008 Till Harbaum <till@harbaum.org>.
 *
 * This file is part of OSM2Go.
 *
 * OSM2Go is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OSM2Go is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OSM2Go.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "appdata.h"

typedef struct {
  appdata_t *appdata;
  GtkWidget *dialog, *view;
  GtkListStore *store;
  GtkWidget *but_add, *but_remove, *but_edit;
  tag_t **tag;
} tag_context_t;

enum {
  TAG_COL_KEY = 0,
  TAG_COL_VALUE,
  TAG_COL_DATA,
  TAG_NUM_COLS
};

static gboolean
view_selection_func(GtkTreeSelection *selection, GtkTreeModel *model,
		     GtkTreePath *path, gboolean path_currently_selected,
		     gpointer userdata) {
  tag_context_t *context = (tag_context_t*)userdata;
  GtkTreeIter iter;
    
  if(gtk_tree_model_get_iter(model, &iter, path)) {
    g_assert(gtk_tree_path_get_depth(path) == 1);

    if(!path_currently_selected) {
      tag_t *tag;
      gtk_tree_model_get(model, &iter, TAG_COL_DATA, &tag, -1);

      /* you just cannot delete or edit the "created_by" tag */
      if(strcasecmp(tag->key, "created_by") == 0) {
	gtk_widget_set_sensitive(context->but_remove, FALSE);
	gtk_widget_set_sensitive(context->but_edit, FALSE);
      } else {
	gtk_widget_set_sensitive(context->but_remove, TRUE);
	gtk_widget_set_sensitive(context->but_edit, TRUE);
      }
    }
  }
  
  return TRUE; /* allow selection state to change */
}

static void on_tag_remove(GtkWidget *but, tag_context_t *context) {
  GtkTreeSelection *selection;
  GtkTreeModel     *model;
  GtkTreeIter       iter;

  selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(context->view));
  if(gtk_tree_selection_get_selected(selection, &model, &iter)) {
    tag_t *tag;
    gtk_tree_model_get(model, &iter, TAG_COL_DATA, &tag, -1);

    g_assert(tag);

    /* de-chain */
    printf("de-chaining tag %s/%s\n", tag->key, tag->value);
    tag_t **prev = context->tag;
    while(*prev != tag) prev = &((*prev)->next);
    *prev = tag->next;

    /* free tag itself */
    gpx_tag_free(tag);

    /* and remove from store */
    gtk_list_store_remove(GTK_LIST_STORE(model), &iter);

    /* but is NULL if remove has been called due to a cancelled add */
    if(but) {
      /* something has been removed -> now dirty */
      map_item_set_flags(&context->appdata->map->selected, GPX_FLAG_DIRTY, 0);
    }
  }
  
  /* disable remove and edit buttons */
  gtk_widget_set_sensitive(context->but_remove,  FALSE);
  gtk_widget_set_sensitive(context->but_edit,  FALSE);
}

static gboolean tag_edit(tag_context_t *context) {

  GtkTreeModel *model;
  GtkTreeIter iter;
  tag_t *tag;

  GtkTreeSelection *sel = gtk_tree_view_get_selection(
		      GTK_TREE_VIEW(context->view));
  if(!sel) {
    printf("got no selection object\n");
    return FALSE;
  }

  if(!gtk_tree_selection_get_selected(sel, &model, &iter)) {
    printf("nothing selected\n");
    return FALSE;
  }
  
  gtk_tree_model_get(model, &iter, TAG_COL_DATA, &tag, -1);
  printf("got %s/%s\n", tag->key, tag->value);

  GtkWidget *dialog = gtk_dialog_new_with_buttons(_("Edit Tag"),
	  GTK_WINDOW(context->dialog), GTK_DIALOG_MODAL,
          GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
	  GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);

  GtkWidget *label, *key, *value;
  GtkWidget *table = gtk_table_new(2, 2, FALSE);

  gtk_table_attach_defaults(GTK_TABLE(table),
		    label = gtk_label_new(_("Key:")), 0, 1, 0, 1);
  gtk_misc_set_alignment(GTK_MISC(label), 1.f, 0.5f);
  gtk_table_attach_defaults(GTK_TABLE(table), 
		    key = gtk_entry_new(), 1, 2, 0, 1);

  gtk_table_attach_defaults(GTK_TABLE(table),
		    label = gtk_label_new(_("Value:")), 0, 1, 1, 2);
  gtk_misc_set_alignment(GTK_MISC(label), 1.f, 0.5f);
  gtk_table_attach_defaults(GTK_TABLE(table), 
		    value = gtk_entry_new(), 1, 2, 1, 2);

  gtk_entry_set_text(GTK_ENTRY(key), tag->key);
  gtk_entry_set_text(GTK_ENTRY(value), tag->value);

  gtk_box_pack_start_defaults(GTK_BOX(GTK_DIALOG(dialog)->vbox), table);

  gtk_widget_show_all(dialog);

  if(GTK_RESPONSE_ACCEPT == gtk_dialog_run(GTK_DIALOG(dialog))) {
    free(tag->key); free(tag->value);
    tag->key = strdup((char*)gtk_entry_get_text(GTK_ENTRY(key)));
    tag->value = strdup((char*)gtk_entry_get_text(GTK_ENTRY(value)));
    printf("setting %s/%s\n", tag->key, tag->value);

    gtk_list_store_set(context->store, &iter,
		       TAG_COL_KEY, tag->key,
		       TAG_COL_VALUE, tag->value,
		       -1);

    gtk_widget_destroy(dialog);
    return TRUE;
  }

  gtk_widget_destroy(dialog);
  return FALSE;
}

static void on_tag_edit(GtkWidget *button, tag_context_t *context) {
  if(tag_edit(context))
    map_item_set_flags(&context->appdata->map->selected, GPX_FLAG_DIRTY, 0);
}

static void on_tag_add(GtkWidget *button, tag_context_t *context) {
  /* search end of tag chain */
  tag_t **tag = context->tag;
  while(*tag) 
    tag = &(*tag)->next;

  /* create and append a new tag */
  *tag = g_new0(tag_t, 1);
  if(!*tag) {
    errorf(GTK_WIDGET(context->appdata->window), _("Out of memory"));
    return;
  }

  /* fill with some dummy data */
  (*tag)->key = strdup(_("<key>"));
  (*tag)->value = strdup(_("<value>"));

  /* append a row for the new data */
  GtkTreeIter iter;
  gtk_list_store_append(context->store, &iter);
  gtk_list_store_set(context->store, &iter,
		     TAG_COL_KEY, (*tag)->key,
		     TAG_COL_VALUE, (*tag)->value,
		     TAG_COL_DATA, *tag,
		     -1);

  gtk_tree_selection_select_iter(gtk_tree_view_get_selection(
		     GTK_TREE_VIEW(context->view)), &iter);

  if(!tag_edit(context)) {
    printf("cancelled\n");
    on_tag_remove(NULL, context);
  } else
    map_item_set_flags(&context->appdata->map->selected, 
		       GPX_FLAG_DIRTY, 0);
}

static GtkWidget *tag_widget(tag_context_t *context) {
  GtkWidget *vbox = gtk_vbox_new(FALSE,3);
  context->view = gtk_tree_view_new();

  gtk_tree_selection_set_select_function(
	 gtk_tree_view_get_selection(GTK_TREE_VIEW(context->view)), 
	 view_selection_func, 
	 context, NULL);

  /* --- "Key" column --- */
  GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
  gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(context->view),
	-1, _("Key"), renderer, "text", TAG_COL_KEY, NULL);

  /* --- "Value" column --- */
  renderer = gtk_cell_renderer_text_new();
  g_object_set(renderer, "ellipsize", PANGO_ELLIPSIZE_END, NULL );
  GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes(
		 _("Value"), renderer, "text", TAG_COL_VALUE, NULL);
  gtk_tree_view_column_set_expand(column, TRUE);
  gtk_tree_view_insert_column(GTK_TREE_VIEW(context->view), column, -1);

  /* build and fill the store */
  context->store = gtk_list_store_new(TAG_NUM_COLS, 
		G_TYPE_STRING, G_TYPE_STRING, G_TYPE_POINTER);

  gtk_tree_view_set_model(GTK_TREE_VIEW(context->view), 
			  GTK_TREE_MODEL(context->store));

  GtkTreeIter iter;
  tag_t *tag = *context->tag;
  while(tag) {
    /* Append a row and fill in some data */
    gtk_list_store_append(context->store, &iter);
    gtk_list_store_set(context->store, &iter,
		       TAG_COL_KEY, tag->key,
		       TAG_COL_VALUE, tag->value,
		       TAG_COL_DATA, tag,
		       -1);
    tag = tag->next;
  }
  
  g_object_unref(context->store);

  /* put it into a scrolled window */
  GtkWidget *scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window), 
				 GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolled_window), 
				      GTK_SHADOW_ETCHED_IN);
  //  gtk_container_set_border_width(GTK_CONTAINER(scrolled_window), 3);
  gtk_container_add(GTK_CONTAINER(scrolled_window), context->view);
  gtk_box_pack_start_defaults(GTK_BOX(vbox), scrolled_window);

  /* ------- button box ------------ */

  GtkWidget *hbox = gtk_hbox_new(TRUE,3);
  context->but_add = gtk_button_new_with_label(_("Add..."));
  gtk_box_pack_start_defaults(GTK_BOX(hbox), context->but_add);
  gtk_signal_connect(GTK_OBJECT(context->but_add), "clicked",
  		     GTK_SIGNAL_FUNC(on_tag_add), context);

  context->but_edit = gtk_button_new_with_label(_("Edit..."));
  gtk_widget_set_sensitive(context->but_edit, FALSE);
  gtk_box_pack_start_defaults(GTK_BOX(hbox), context->but_edit);
  gtk_signal_connect(GTK_OBJECT(context->but_edit), "clicked",
  		     GTK_SIGNAL_FUNC(on_tag_edit), context);

  context->but_remove = gtk_button_new_with_label(_("Remove"));
  gtk_widget_set_sensitive(context->but_remove, FALSE);
  gtk_box_pack_start_defaults(GTK_BOX(hbox), context->but_remove);
  gtk_signal_connect(GTK_OBJECT(context->but_remove), "clicked",
		     GTK_SIGNAL_FUNC(on_tag_remove), context);

  gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
  return vbox;
}

void info_dialog(appdata_t *appdata) {
  g_assert(appdata->map->selected.type != MAP_TYPE_ILLEGAL);
  tag_context_t *context = g_new0(tag_context_t, 1);
  user_t *user = NULL;
  char *str = NULL;
  time_t stime = 0;

  context->appdata = appdata;

  switch(appdata->map->selected.type) {
  case MAP_TYPE_NODE:
    str = g_strdup_printf(_("Node #%ld"), appdata->map->selected.node->id);
    user = appdata->map->selected.node->user;
    context->tag = &appdata->map->selected.node->tag;
    stime = appdata->map->selected.node->time;
    break;
  case MAP_TYPE_WAY:
    str = g_strdup_printf(_("Way #%ld"), appdata->map->selected.way->id);
    user = appdata->map->selected.way->user;
    context->tag = &appdata->map->selected.way->tag;
    stime = appdata->map->selected.way->time;
    break;
  default:
    g_assert((appdata->map->selected.type == MAP_TYPE_NODE) ||
	     (appdata->map->selected.type == MAP_TYPE_WAY));
    break;
  }

  context->dialog = gtk_dialog_new_with_buttons(str,
	  GTK_WINDOW(appdata->window), GTK_DIALOG_MODAL,
	  GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE, NULL);
  g_free(str);

  /* making the dialog a little wider makes it less "crowded" */
#ifdef USE_HILDON
  gtk_window_set_default_size(GTK_WINDOW(context->dialog), 500, 300);
#else
  gtk_window_set_default_size(GTK_WINDOW(context->dialog), 400, 200);
#endif

  GtkWidget *label;
  GtkWidget *table = gtk_table_new(2, 2, FALSE);  // x, y

  /* ------------ user ----------------- */
  label = gtk_label_new(_("User:"));
  gtk_table_attach_defaults(GTK_TABLE(table),  label, 0, 1, 0, 1);
  if(user) label = gtk_label_new(user->name);
  else     label = gtk_label_new("---");
  gtk_table_attach_defaults(GTK_TABLE(table),  label, 1, 2, 0, 1);

  /* ------------ time ----------------- */
  label = gtk_label_new(_("Time:"));
  gtk_table_attach_defaults(GTK_TABLE(table),  label, 0, 1, 1, 2);
  struct tm *loctime = localtime(&stime);
  char time_str[32];
  strftime(time_str, sizeof(time_str), "%x %X", loctime);
  label = gtk_label_new(time_str);
  gtk_table_attach_defaults(GTK_TABLE(table),  label, 1, 2, 1, 2);

  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(context->dialog)->vbox), table, 
		     FALSE, FALSE, 0);

  /* ------------ tags ----------------- */

  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(context->dialog)->vbox),
		     tag_widget(context), TRUE, TRUE, 0);

  /* ----------------------------------- */

  gtk_widget_show_all(context->dialog);
  gtk_dialog_run(GTK_DIALOG(context->dialog));
  gtk_widget_destroy(context->dialog);

  /* since nodes being parts of ways but with no tags are invisible, */
  /* the result of editing them may have changed their visibility */
  map_item_redraw(appdata, &appdata->map->selected);
}
